import torch
import pandas as pd
import numpy as np
import cv2
import os
from albumentations import (
    Compose, Resize, Normalize, RandomBrightnessContrast,
    HorizontalFlip, VerticalFlip, Rotate, GaussianBlur
)
from albumentations.pytorch import ToTensorV2
from torch import nn
# Import the model and config from your existing code
from aiia import AIIA, AIIAConfig, AIIABase, AIIABaseShared, AIIAmoe, AIIAchunked, AIIArecursive

class aiuNNDataset(torch.utils.data.Dataset):
    def __init__(self, parquet_path, config=None):       
        # Read the Parquet file
        self.df = pd.read_parquet(parquet_path)
        
        # Data augmentation pipeline
        self.augmentation = Compose([
            Resize(height=512, width=512),
            RandomBrightnessContrast(),
            HorizontalFlip(p=0.5),
            VerticalFlip(p=0.5),
            Rotate(limit=45),
            GaussianBlur(p=0.3),
            Normalize(mean=[0.5], std=[0.5]),
            ToTensorV2()
        ])
        
    def __len__(self):
        return len(self.df)
    
    def __getitem__(self, idx):
        # Get the byte strings
        low_res_bytes = self.df.iloc[idx]['low_res']
        high_res_bytes = self.df.iloc[idx]['high_res']
        
        # Convert bytes to numpy arrays
        low_res = cv2.imdecode(np.frombuffer(low_res_bytes, np.uint8), -1)
        high_res = cv2.imdecode(np.frombuffer(high_res_bytes, np.uint8), -1)
        
        # Apply augmentation and normalization
        augmented = self.augmentation(image=low_res, mask=high_res)
        low_res = augmented['image']
        high_res = augmented['mask']
        
        return {
            'low_res': low_res,
            'high_res': high_res
        }


def finetune_model(model: AIIA, train_parquet_path, val_parquet_path, batch_size=8, epochs = 10):
    # Initialize dataset and dataloader
    train_dataset = aiuNNDataset(train_parquet_path)
    val_dataset = aiuNNDataset(val_parquet_path)
    
    train_loader = torch.utils.data.DataLoader(
        train_dataset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=4
    )
    
    val_loader = torch.utils.data.DataLoader(
        val_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=4
    )
    
    # Set device
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    model.to(device)
    
    # Define loss function and optimizer
    criterion = nn.MSELoss()
    optimizer = torch.optim.Adam(model.parameters(), lr=model.config.learning_rate)
    
    best_val_loss = float('inf')
    
    for epoch in range(epochs):
        model.train()
        
        train_loss = 0.0
        
        for batch_idx, batch in enumerate(train_loader):
            low_res = batch['low_res'].to(device)
            high_res = batch['high_res'].to(device)
            
            # Forward pass
            outputs = model(low_res)
            
            # Calculate loss
            loss = criterion(outputs, high_res.permute(0, 3, 1, 2))  # Adjust for channel dimensions
            
            # Backward pass and optimize
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            
            train_loss += loss.item()
            
        avg_train_loss = train_loss / len(train_loader)
        
        print(f"Epoch {epoch+1}, Training Loss: {avg_train_loss:.4f}")
        
        # Validation
        model.eval()
        val_loss = 0.0
        
        with torch.no_grad():
            for batch in val_loader:
                low_res = batch['low_res'].to(device)
                high_res = batch['high_res'].to(device)
                
                outputs = model(low_res)
                loss = criterion(outputs, high_res.permute(0, 3, 1, 2))
                
                val_loss += loss.item()
        
        avg_val_loss = val_loss / len(val_loader)
        
        print(f"Epoch {epoch+1}, Validation Loss: {avg_val_loss:.4f}")
        
        # Save best model
        if avg_val_loss < best_val_loss:
            best_val_loss = avg_val_loss
            model.save("best_model")
            
    return model

def main():
    # Paths to your data
    train_parquet_path = "/root/training_data/vision-dataset/image_upscaler.parquet"
    val_parquet_path = "/root/training_data/vision-dataset/image_vec_upscaler.parquet"
                
    # Load pretrained model
    model = AIIA.load("/root/vision/AIIA/AIIA-base-512")
    
    # Add final upsampling layer if needed (depending on your specific architecture)
    if hasattr(model, 'chunked_'):
        model.add_module('final_upsample', nn.Upsample(scale_factor=2, mode='bilinear'))
        
    # Fine-tune
    finetune_model(
        model,
        train_parquet_path,
        val_parquet_path
    )
    
if __name__ == '__main__':
    main()